#include "forge.h"

#ifndef TESTIFY_H
#define TESTIFY_H
/*  -------------- Testify --------------------
Testify is a library for sereielizing data in to bit streams, and then unpacking them. Data can be packed/unpacked form memory, a network connection or a file. Testify packs all data in big endian order and hides any difference between big or small endianess of the CPU. While Testify is written to be very fast it also has extensive debugging capabillities. By enabling the debug state of a Stream Testify can be made to pack significant ammounts of meta data arround each entry. By doing this the receving end can match the incomming data with the expected data makig it trivial to find pack/unpack miss-match buggs. By building Testify without the define TESTIFY_DEBUG all debugging code can be dissabled and maximum preformance can be obtained.*/


#define TESTIFY_IS_BIG_ENDIAN (*(short *)"\0\xff" < 0x100)

#define U_BUFFER_SIZE 2048

#define TESTIFY_HANDLE_ADDRESS_COUNT_MAX 16

#define TESTIFY_DEBUG

typedef enum{
	TESTIFY_TRANSPORT_NULL,
	TESTIFY_TRANSPORT_IPV4,
	TESTIFY_TRANSPORT_IPV6,
	TESTIFY_TRANSPORT_BLUETOOTH,
	TESTIFY_TRANSPORT_COUNT
}TestifyTransport;

typedef enum{
	TESTIFY_PROTOCOL_DATAGRAM,
	TESTIFY_PROTOCOL_STREAM,
	TESTIFY_PROTOCOL_RAW
}TestifyProtocol;

typedef enum{
	TESTIFY_AT_LOOPBACK,
	TESTIFY_AT_PRIVATE_NETWORK,
	TESTIFY_AT_DOCUMENTATION,
	TESTIFY_AT_BROADCAST,
	TESTIFY_AT_MULTICAST,
	TESTIFY_AT_GLOBAL
}TestifyAddressType;

typedef enum{
	TESTIFY_FLAG_BIND = 0x1,
	TESTIFY_FLAG_BLOCKING = 0x2,
	TESTIFY_FLAG_PORT_MAPPED = 0x4,
	TESTIFY_FLAG_NON_BLOCKING_CONNECT = 0x8
}TestifyFlags;

typedef struct{
	TestifyTransport transport;
	union{
		uint32 ipv4;
		uint16 ipv6[128 / 16];
		uint8 bluetooth[6];
	}address;
	uint16 port;
}TestifyAddress;

#define TESTIFY_NETWORK_PEER_MESSAGE_LENGTH 200

typedef enum{
	TESTIFY_SST_ADAPTER, // IP Address of an installed adapter
	TESTIFY_SST_GATEWAY, // IP Address of a gateway
	TESTIFY_SST_BROADCAST, // IP Broadcast address;
	TESTIFY_SST_MAPPED_ADDRESS, // IP address used by router;
	TESTIFY_SST_DNS_GIVEN, // DNS Address configured on the system.
	TESTIFY_SST_DNS, // Alternative DNS addresses
	TESTIFY_SST_DNS_VERIFY, // Addresses used to verify that DNS works.
	TESTIFY_SST_HTTP, // Http servers that return the global ip of the computer
	TESTIFY_SST_IRC, // DNS server to known IRC servers.
	TESTIFY_SST_MASTER_SERVER, // Addresses to Master servers.
	TESTIFY_SST_COUNT
}TestifyServiceType;

typedef enum{
	TESTIFY_SS_INACTIVE, /* keep the data until futher notice*/
	TESTIFY_SS_TRIGGERD, /* do something with this one */
	TESTIFY_SS_WAITING,
	TESTIFY_SS_SUCSESS,
	TESTIFY_SS_SUCSESS_DOWNGRADED, /* older protocol */
	TESTIFY_SS_FAILED,
	TESTIFY_SS_COUNT
}TestifyServiceState;

typedef enum{
	TESTIFY_NS_CONNECTED					= 0x1, // Do we have any IP network connection at all.
	TESTIFY_NS_IPV4							= 0x2, // Do we have a IPV4 Address
	TESTIFY_NS_IPV6							= 0x4, // Do we have a IPV6 Address
	TESTIFY_NS_DNS							= 0x8, // Do we have DNS access
	TESTIFY_NS_INTERNET_ACCESS_CONFIRMED	= 0x10, // Do we have confirmed internet access
	TESTIFY_NS_LAN_ACCESS					= 0x20, // Are we on a Lan
	TESTIFY_NS_GLOBAL_IP_V4					= 0x40, // Does this machin have aglobal IP
	TESTIFY_NS_GLOBAL_IP_V6					= 0x80, // Does this machin have aglobal IP
	TESTIFY_NS_GLOBAL_IP_CONFIRMED			= 0x100, // Is our global or NAT global IP Confirmed
	TESTIFY_NS_BEHIND_NAT					= 0x200, // Are we behind a NAT
	TESTIFY_NS_BEHIND_LOGIN					= 0x400, // Are we behind a NAT
	TESTIFY_NS_PORT_FORWARDING_AVAILABLE	= 0x800, // Is any form of port forwarding available
	TESTIFY_NS_PORTMAPED_PCP				= 0x1000, // The nat we are behind supports PCP
	TESTIFY_NS_PORTMAPED_PMP				= 0x2000, // The nat we are behind supports PMP
	TESTIFY_NS_COUNT						= 14 // Number of flags
}TestifyNetworkStatus;

typedef enum{
	TESTIFY_DS_NONE,
	TESTIFY_DS_ADDED,
	TESTIFY_DS_REMOVED,
	TESTIFY_DS_COUNT
}TDiscoverDiscoverState;

typedef enum{
	TESTIFY_DMF_BROADCAST_LOCAL = 0x1,
	TESTIFY_DMF_BROADCAST_LAN = 0x2,
	TESTIFY_DMF_BROADCAST_WAN = 0x4,
	TESTIFY_DMF_WAN_IRC = 0x8,
	TESTIFY_DMF_BUETOOTH = 0x10
}TDiscoverDiscoverMethodFlag;



typedef struct{
	TestifyAddress address;
	char message[TESTIFY_NETWORK_PEER_MESSAGE_LENGTH];
	uint32 method;
}TestifyDiscoverPeer;


/* ---------- TCP Networking --------
Handles the creation and mangement of TCP streams.*/


//#ifdef TESTIFY_INTERNAL
#include "t_internal.h"
//#else
//typedef void TestifyHandle;
//typedef void TestifyDiscovery;
//#endif

//#define TESTIFY_DEPRICATE_LEGACY_API
#ifndef TESTIFY_DEPRICATE_LEGACY_API
#include "testify_legacy.h"
#endif


boolean testify_address_lookup(TestifyAddress *address, char *dns_name, TestifyTransport transport, uint16 default_port);

extern TestifyHandle	*testify_create(TestifyAddress *address, TestifyProtocol protocol, uint64 flags); /* Creates a TCP connection that can be both read or written to, connecting to the DNS address given as "host_name". By default the connection will attempt to connect to "port", unless the host_name is frormated in the following way <dns name>:<port>, then the port number will be read from the host_name. If no host name is given, the function will open up the port, and wait for incomming connections.*/
extern void				testify_destroy(TestifyHandle *handle); /* frees a handle */
extern TestifyHandle	*testify_wait_for_connection(TestifyHandle *listener); /* Waits for an incomming TCP or BLUETOOTH connection */
extern boolean			testify_address(TestifyHandle *handle, uint id, TestifyAddress *output); /* Returns the address of a Testify handle. A handle can have up to TESTIFY_HANDLE_ADDRESS_COUNT_MAX number of addresses */
extern TestifyProtocol	testify_protocol(TestifyHandle *handle); /* Returns the type of a Testify handle. */
extern boolean			testify_connected(TestifyHandle *handle); /* returns true if the handle is still connected */

#define TESTIFY_ADDRESS_MAX_STRING_SIZE 47
extern void				testify_address_print(TestifyAddress *address, char *buffer); /* Prints out an address. buffer needs to be at least TESTIFY_ADDRESS_MAX_STRING_SIZE bytes long.*/
extern boolean			testify_address_compare(TestifyAddress *a, TestifyAddress *b); /* compares two address structures */
extern boolean			testify_address_compare_without_port(TestifyAddress *a, TestifyAddress *b);
extern TestifyAddressType testify_address_type(TestifyAddress *address); /* Categorizes an Address*/

#define TESTIFY_ADDRESS_MAX_SERIALIZE_SIZE (3 + 8 * 2)
extern size_t			testify_address_serialize(TestifyAddress *address, uint8 *buffer);
extern size_t			testify_address_deserialize(TestifyAddress *address, uint8 *buffer, uint buffer_length);

extern size_t			testify_stream_send(TestifyHandle *handle, uint8 *buffer, size_t buffer_size);
extern size_t			testify_stream_send_progressivly(TestifyHandle *handle, uint8 *buffer, size_t buffer_size);
extern size_t			testify_stream_recive(TestifyHandle *handle, uint8 *buffer, size_t buffer_size);
extern size_t			testify_datagram_send(TestifyHandle *handle, TestifyAddress *to, uint8 *buffer, size_t size);
extern size_t			testify_datagram_receive(TestifyHandle *handle, TestifyAddress *from, uint8 *buffer, size_t size);

extern void				testify_pack_bigendian16(void *output, void *input); /* Packing 16 from machine order to big endian order. */
extern void				testify_pack_bigendian32(void *output, void *input); /* Packing 32 from machine order to big endian order. */
extern void				testify_pack_bigendian64(void *output, void *input); /* Packing 64 from machine order to big endian order. */

extern void				testify_unpack_bigendian16(void *output, void *input); /* Unpacking 16 from big endian order to machine order. */
extern void				testify_unpack_bigendian32(void *output, void *input); /* Packing 32 from big endian order to machine order. */
extern void				testify_unpack_bigendian64(void *output, void *input); /* Packing 64 from big endian order to machine order. */

extern void				testify_pack_littleendian16(void *output, void *input); /* Packing 16 from machine order to little endian order. */
extern void				testify_pack_littleendian32(void *output, void *input); /* Packing 32 from machine order to little endian order. */
extern void				testify_pack_littleendian64(void *output, void *input); /* Packing 64 from machine order to little endian order. */

extern void				testify_unpack_littleendian16(void *output, void *input); /* Unpacking 16 from little endian order to machine order. */
extern void				testify_unpack_littleendian32(void *output, void *input); /* Packing 32 from little endian order to machine order. */
extern void				testify_unpack_littleendian64(void *output, void *input); /* Packing 64 from little endian order to machine order. */

#include "t_pack.h"

/* Discovery system */

extern TestifyDiscovery	*testify_discover_create(char *service, TestifyHandle *handle, char *message, uint32 method_flags);
extern void				testify_discover_destroy(TestifyDiscovery *discover);
extern boolean			testify_discover_updated(TestifyDiscovery *discover);
extern TestifyDiscoverPeer	*testify_discover_service_list_get(TestifyDiscovery *discover, uint *peer_count);

/* Network status */

extern uint64			testify_status_flags_get(void); /* currently set flags */
extern void				testify_status_debug_print(); /* print the Current state to stdout */
extern void				testify_status_initialize_default(boolean activate); /* initializes all different services with built in settings. If activated is TRUE, all these services will be activated. Only activate when debugging, or diagonsing the network. */
extern void				testify_status_service_set(TestifyServiceType type, char *name, TestifyAddress *address);
extern boolean			testify_status_service_get(uint nr, TestifyServiceType type, char *name, uint buffer_size, TestifyAddress *address, TestifyServiceState *status);

/* IRC implementation */

typedef enum{
	TESTIFY_IRC_COMMAND_JOIN = 1000,
	TESTIFY_IRC_COMMAND_PART,
	TESTIFY_IRC_COMMAND_PRIVMSG,
	TESTIFY_IRC_COMMAND_NOTICE ,
	TESTIFY_IRC_COMMAND_NICK,
	TESTIFY_IRC_COMMAND_NAMES,
	TESTIFY_IRC_COMMAND_PING,
	TESTIFY_IRC_COMMAND_QUIT,
	TESTIFY_IRC_COMMAND_SERVLIST,
	TESTIFY_IRC_COMMAND_TOPIC,
	TESTIFY_IRC_COMMAND_USERS,
	TESTIFY_IRC_COMMAND_OPER,
}TestifyIrcCommands;

TestifyHandle			*testify_irc_create(char *server, char *user_name, char *channel);
extern void				testify_irc_channel_join(TestifyHandle *handle, char *channel_name);
TestifyHandle			*testify_irc_parser(TestifyHandle *handle, void (*command_func)(uint command, char *sender, char *channel, char *mesassage, void *user), void *user);
extern void				testify_irc_nick_set(TestifyHandle *handle, char *nick);
extern char				*testify_irc_nick_get(TestifyHandle *handle);

extern uint				testify_irc_channel_count_get(TestifyHandle *handle);
extern char				*testify_irc_channel_name_get(TestifyHandle *handle, uint channel);
extern uint				testify_irc_channel_user_count_get(TestifyHandle *handle, uint channel);
extern char				*testify_irc_channel_user_name_get(TestifyHandle *handle, uint channel, uint user_id);
extern uint				testify_irc_channel_message_count_get(TestifyHandle *handle, uint channel);
extern char				*testify_irc_channel_message_get(TestifyHandle *handle, uint channel, uint message_id);
extern char				*testify_irc_channel_message_user_name_get(TestifyHandle *handle, uint channel, uint message_id);
extern void				testify_irc_channel_message_send(TestifyHandle *handle, uint channel, char *message);

/*
*

void function(unsigned long long *data, unsigned long long *arg, unsigned long long i)
{
	unsigned long long key, rotate;
	key = arg[i & 15];
	data[i] ^= arg[key & 15] ^ arg[((~key) & 1) | ((key >> 4) & 14)];
	rotate = (key << 1) | ((key >> 63) & 1);
	arg[i & 15] = rotate;
	arg[key & 15] ^= rotate;
}
	*/

#endif